/****************************************************************************
 *                                                                          *
 * This program called cvi (C verbal interpreter) looks for information     *
 * about variables and data types in an ANSI C program.                     *
 * Copyright (c) 1999 Chepelianskii Alexeii.                                *
 *                                                                          *
 * This program is free software; you can redistribute it and/or modify     *
 * it under the terms of the GNU General Public License as Published by     *
 * the Free Software Fondation; either version 2 of the License, or (at     *
 * your option) any later version.                                          *
 *                                                                          *
 * This program is distributed in the hope it will be useful, but WITHOUT   *
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or    *
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License     *
 * for more details.                                                        *
 *                                                                          *
 * You should have received a copy of the GNU General Public License along  *
 * with this program; if not, write to the Free Software Foundation, Inc.,  *
 * 675 Mass Ave, Cambridge, MA 02139, USA.                                  *
 *                                                                          *
 ****************************************************************************/
 /***************************************************************************  

   Informations about the algorithms used here can be found in
   Programs and Data structures in C by Leendert Ammeraal.
   The C Programming Language by B. Kernighan & D. Ritchie. (2nd edition).

   C grammar and syntax sources are :
   The C Programming Language.
   The C++ Programming Language by B. Stroustrup
   (in several cases C code is parsed as if it were an ANSI C++ code).
   The gcc info-manual pages.
   
   The parsed program is presumed to be wrote using a correct syntax
   otherwise (particularly for the old C style I do not know well)
   it may generate wrong results instead of errors messages and even
   worst : you were warned.

   However I wrote this software to be useful, and I hope it is, even in
   its present form.

 ****************************************************************************/
 /***************************************************************************

   note : _Although it is an ANSI C program it may be compiled as a C++
   program it that case inline functions will be used.
   (They are also used when __GNUC__ is defined ).
          _This program was originally written in French so that often
   names of functions and variables are not derived from English words.

 ****************************************************************************/

#if !defined __GNUC__ && !defined __cplusplus
      #define inline          /* empty world */
#endif

#ifndef STACK
#define STACK

/*************************************************/
/* First of all, a version of a generic C stack. */
/*************************************************/
typedef struct List list;

void stackInitG(list **l);          /* can be replaced by *l = NULL */
void *pushG(int size, list **);
void *getG(list *);
void popG(list **);
void cleanG(list **);
void echG(list *);
int howManyG(list *);
int isElemG(list *);
int setToFirstG(list **);
int setToLastG(list **);
int setToNextG(list **);
int setToPrevG(list **);
#endif


/*****************************************************/
/* All data types and functions to parse a C program */
/* are here, to analyze a variable definition or to  */
/* find out how it is used further in the program.   */
/*****************************************************/

#ifndef TRAD
#define TRAD

/*****************************************************/
/* I) constatnts that may often be changed .         */
/*****************************************************/

/* To distinguish predefined name and user ones, a forbidden
   identification letter is used.
*/

enum { SPECIAL = '@'};
#define EXTERN "@the_external_scope"
#define DEFAULT "@default_named_"
#define USER "@user"
#define ENUMT "@enumType "
#define INPUT "@input"
#define GENER "@generic"

/*****************************************************/
/* II): Data types to organize resulting information.*/
/*****************************************************/

struct variable;
struct Struc;

/* Store information about a variable. */
struct infovar {       /** Strings that serve to define a variable */
    char *nom;         /* its name  */
    char *dec;         /* ugly definition text for type verification */
    char *text;        /* the output readable text */
    char *type;        /* its reorganized type definition */
    char *strtype;     /* its real type */
    int whatIs;        /* area, function, pointer ... */
    struct Struc *str; /* used for the -m option */
                       /** Information about the scope of the variable */
    int status;        /* Is it in the current bloc ? */
    int def;           /* == 0, if the variable is only declared. */
                       /** To describe where the variable is */
    char *file;        /* file where the variable is definied. */
    char *fonction;    /* function where the variable is defined. */
    int ligne;         /* line of the definition */
};

/* This is a structure to store variables in a binary tree. */
typedef struct variable {
    int ntrouves;
    struct infovar var;
    struct variable *gauche;
    struct variable *droit;
    struct variable *copyOfStr;    /* Copy of the car.str structure */
    char *copyName;                /* name of the membersOf entry */
} Variable;

/* As macros can modify the way variables are defined,
   a miniature version of the C preprocessor is implemented.
   However this one is really limited and does not know
   anything about conditional compilation. */
/* for macro parsing : structure which contains macro arguments */
typedef struct Args {
    char *name;                    /* name of the argument */
    char *replace;                 /* value of the argument */
    int n;
    struct Args *next;
} Args;

/* Here is the support for macros, structures typedefs and enumerations */
typedef struct Struc {
    struct infovar str;
    union {
        Args **args;               /* for preprocessor constants */
        Variable *dec;             /* for structures/unions */
        int *type;                 /* for typedef */
    } obj;
    struct Struc *suiv;            /* to form a list */
} Struc;

/* these constants are used to distinguish different fields of the 
   Struc::obj union. */
typedef enum entry { STR_INF,      /* structures/unions */
		     TP_INF,       /* typedefs */
		     ENUM_INF ,    /* enumerations */
		     MACRO_INF     /* macros */} 
Entry;

/* that structure explains how the collected information
   should be printed */
typedef struct condition {
    Variable *var;                 /* binary tree of variables */
    Variable *bloc;                /* binary tree of blocs */
    Variable *str;                 /* binary tree of structures to parse */
    Variable *Affstruct;
    Variable *Afftypedef;
    Variable *Affenum;
    Variable *Affmacro;
    int affstruct;
    int afftypedef;
    int affenum;
    int affmacro;
    int noaffvar;
    int affmax;                    /* request for very used variables */
    int affmin;                    /* to detect unused variables */
    list *files;
} Condition;

/* This type is used to store information about the program state
   in a stack, as recursive functions are often useed. */
typedef struct {
    Variable **var;
    char *func; 
} Prog;

/* As bsearch is used to avoid duplication of code that type is useful. */
typedef int (*ARG)(const void *, const void *);


/****************************************************/
/* III) : functions to manipulate these data types. */
/****************************************************/

/** Some of the following functions are directly used when parsing a
   program. Others were written for a more general purpose. */

/** General purpose algorithms. */

/* functions to manipulate the input stream. */
int lirecar(void);                   /* to read a character from the input */
void remettrecar(int charactere);    /* to put a letter back to the input */
void remettrech(char *);             /* the same thing for a string */
int lirewspace(void);                /* to skip spaces */
int negligerComment(int letter);     /* to skip quoted text */
int negligerGuillemets(int);         /* to skip C style comments */
char *getQuote(void);                /* to restore a quote skipped by */
                                     /* negligerGuillemets */
char *conv(int nombre);              /* to convert an integer to a string */

/* functions that operate on words. */
int readword(char **mot , int *maxmot);
int liremot(char **mot, int *maxmot);
int motcmp(char *s1, char *s2);

/* functions that operate on dynamic allocated strings. */
char *dupchaine(char *chaine);
char *mstrcat(char **dest, const char *src);
char *mstrcpy(char **dest, const char *src);
char *vastrcat(int narg, char **format, ...);
char *mistrcat(char **dest, const char *src, int *size);
char *mistrcpy(char **dest, const char *src, int *size);
void maketable(int ncolumns, int leftmargin, int rightmargin, ...);

/* error handing */
void error(char *ftp, ...);
void fatalerror(char *ftp, ...);
void AllocError(void);
int iserror(void);

/* to copy variables lists from a structure to another. */
void copierVar(Variable **, Variable *);

/** Algorithms to parse C programs */

/* To know if a token is a C type */
int dicho(char *nom, const char **liste, int dimliste);
struct Type *dichoType(char *nom, const struct Type *, int dim);
int IsType(char *);

/* trad is a function that reads a C declaration and translate it in the 
   corresponding infovar structure, it returns an integer to indicate 
   what kind of variable was found : function definition ...
*/
int trad(Variable **var, struct infovar *NomVar, char *CurrentFunc);

/* Then trad put the new variable in the *var structure which contains
   the list of all the variables the parsed program uses, to do so it calls */
Variable *ajoutarbre(Variable *racine, struct infovar *var);

/* Further others part of the program will search information about variables,
   types ... */
/* So to find information about a type, an enumeration  */
unsigned int hacher(const char *, int lim);
Struc *consulter(char *name, Entry);
/* are used. */

/* In the same time */
int unknownvar(Variable **racine, char *func, char **mot, int *sz);
int anaind(Variable **racine, char *func, char **mot, int *sz);
Variable *trouverDansArbre(Variable *racine, char *mot, char *func);
Variable *StrictTrouver(Variable *racine, char *mot, char *func);
/* are used to find which variable a token represents. */

/* To put variables out of scope once a bloc is finished */
void newbloc(Variable *);
void clean(void);
/* functions are provided they are important for a correct understanding
   of a program. */


/* At the end of the execution that functions are used to display the
   collected information. */
void dispType(const Condition *);
void affDefStruc(const Condition *);
void affDefEnum(const Condition *);
void affDefMacro(const Condition *);
void affarbre(Variable *racine, const Condition *);
void affDefMembers(Variable *racine);

int findMacro(char *, char *);
int OpenMacro(char *);

/** Constants and external variables to keep track of the program state
    and to communicate information between different functions. */

/* information about C keywords. */
extern const struct Type donnee[];  /* types : 'int', 'float' */
extern const char *motcle[];        /* keywords of the C language.*/
extern int const NCLES;             /* number of elements in donnee */
extern int const NREST;             /*        ,,,         in  motcle */

/* Then many parsing operations require identification of tokens. varialbes
   states ... */
/* Identification of a token */
enum { ERREUR      ,     /* an error occurred */
       DECLARATION ,     /* definition of a variable */
       MENTION     ,     /* definition of a function */
       CAST        ,     /* a cast */
       DEFSTRUC    ,     /* structure definition */
       DEFENUM     ,     /* enumeration definition */
       TYPEDEF     ,     /* typedef statment */
       OLD_MENTION       /* old styled function definition */
     } typedef UserTypes;
/* variable state */
enum { RESERVED = -2,    /* Declared but not defined yet */
       OLD = -1          /* definitively out of scope */ };
/* description of variable types */
enum { UNDEF    ,        /* Nothing */ 
       POINTEUR ,        /* pointer */
       VAR      ,        /* a baisc type variable */
       TABLEAU  ,        /* area */
       FONCTION          /* function */ };
/* other constants which are not as used */
enum { STRUCTURE, UNION };

/* general state of the program */

extern int nlignes;        /* number of lines */
extern int nblocs;         /* number of opened '{' */
extern char *filename;     /* name of the current file */

#endif

/***********************************************************/
/* Parsing a type is in fact a complicate operation since  */
/* it can contain structures, enumerations, types          */
/*   definitions hence this part of thee header file.      */
/* It is used in most cases by trad and children routines. */
/***********************************************************/
#ifndef ANA
#define ANA

/* status of different type specifiers, acces modifiers ... */
enum {  INDULGENT = -1,  /* no type verificaton */
	QUAL   ,         /* long short : basic data type modifiers */
	STOCK  ,         /* classe de stockage : static ...*/
        STRUCT ,         /* structures, unions */ 
	TYPE   ,         /* int , double ... */
	ACCSMOD,         /* access modifiers : const, volatile */ 
	ENUM             /* enumerations */
       } typedef TypeOfVar;

/* To ensure types will be compared correctly they are converted in an area
   of integers that describe them : this structure is used to associate
   a TypeOfVar to a basic data type like int, ... */
struct Type { 
    char *nomtype;
    TypeOfVar status;
    int ntrouve;
};              

/* While parsing, information is not directly stored in a Variable structure,
   instead this one is employed to store information which will be useless
   further.
*/
typedef struct Sortie {
    int erreurpos;          /* error possible */
    int quelvar[3];         /* type of the two last and of the first token */
    int arg;                /* for a function, number of its arguments */
    struct infovar var;     /* 'var' will be copied in the variable list */
} Sortie;
enum { LAST, MIDDLE, FIRST };

/* useful constants */
/* low level token identification */
enum { NOM , PARENS , INDEFPARENS , CROCHETS, OLDPARENS};
enum CTypes 
     { User, Auto, Char, Const, Double, Enum, Extern, Float, Int, Long, 
       Register, Short, Signed, Static, Struct, Union, Unsigned ,Void , 
       Volatile};

/* others */
enum { DECL = 1 }; 
enum { NON, OUI };

/* Moreover lex and typelex are used to communicate easily tokens between
   functions. */
extern char *lex;                /* the last token */
extern int typelex;              /* its low level type : PARENS ... */

/* Routines to help trad to parse a C type, and to define a new type
   when typedef is used.
*/
int liretype(int niv, struct Type *, int inf, struct Sortie *out, char *);
Struc *defineType(struct Sortie *, char *bloc, struct Type *, int nelem, 
   int status);

/* Others output-input functions used by trad() and its children routines. */
int lirelex(void);
void remettre(int typelex);
#endif

